/*
    Created from:
        Messages = <INPUTFILENAME>
        Template = <TEMPLATEFILENAME>
        Language = <LANGUAGEFILENAME>

                     AUTOGENERATED FILE, DO NOT EDIT

*/
#ifndef <MSGNAME>HeaderStruct_H__
#define <MSGNAME>HeaderStruct_H__

//# What does Simulink want for message headers, if anything?!?

#endif
