/*
    <OUTPUTFILENAME>
    Created <DATE> from:
        Messages = <INPUTFILENAME>
        Template = <TEMPLATEFILENAME>
        Language = <LANGUAGEFILENAME>

                     AUTOGENERATED FILE, DO NOT EDIT

*/
package <MESSAGE_PACKAGE>

import java.nio.ByteBuffer

class <MSGSHORTNAME> {
    companion object {
        const val SIZE = <MSGSIZE>
    }

    var data: ByteBuffer

    constructor() {
        data = ByteBuffer.allocate(SIZE)
        initialize()
    }

    constructor(buffer: ByteBuffer) {
        data = buffer
    }

    fun initialize() {
        <INIT_CODE>
    }

    <ENUMERATIONS>
    <FIELDINFOS>
    fun setMessageId(id: UInt) {
        <SETMSGID>
    }

    fun getMessageId(): UInt {
        <GETMSGID>
    }

    <ACCESSORS>
}
