/*
    <OUTPUTFILENAME>
    Created <DATE> from:
        Messages = <INPUTFILENAME>
        Template = <TEMPLATEFILENAME>
        Language = <LANGUAGEFILENAME>

                     AUTOGENERATED FILE, DO NOT EDIT

*/
package <MESSAGE_PACKAGE>;

import java.nio.ByteBuffer;
import java.util.HashMap;
import java.util.Map;
import msgtools.FieldAccess;

public class <MSGNAME>
{
    public static final int SIZE = <MSGSIZE>;
    public <MSGNAME>()
    {
        m_data = ByteBuffer.allocate(SIZE);
        Init();
    }
    public <MSGNAME>(ByteBuffer buffer)
    {
        m_data = ByteBuffer.wrap( buffer.array(), 0, SIZE );
    }
    public ByteBuffer GetBuffer() 
    { 
        return m_data; 
    }
    void Init()
    {
        <INIT_CODE>
    }
    <ENUMERATIONS>
    <FIELDINFOS>
    public void SetMessageID(int id)
    {
        <SETMSGID>;
    }
    public int GetMessageID()
    {
        return (int)<GETMSGID>;
    }
    <ACCESSORS>
    private ByteBuffer m_data;
};
