/*
    <OUTPUTFILENAME>
    Created <DATE> from:
        Messages = <INPUTFILENAME>
        Template = <TEMPLATEFILENAME>
        Language = <LANGUAGEFILENAME>

                     AUTOGENERATED FILE, DO NOT EDIT

*/
import Foundation

public class <MSGNAME>
{
    public static let MSG_ID = UInt32(<MSGID>);
    public static let MSG_SIZE = <MSGSIZE>;
    public static let MSG_NAME = "<MSGDESCRIPTOR>";
    var hdr: NetworkHeader;
    var m_data: NSMutableData;
    init()
    {
        m_data = NSMutableData(length: NetworkHeader.MSG_SIZE + <MSGNAME>.MSG_SIZE)!;
        hdr = NetworkHeader(buffer: m_data);
        hdr.SetMessageID(<MSGNAME>.MSG_ID);
        hdr.SetDataLength(UInt32(<MSGNAME>.MSG_SIZE));
        //InitializeTime();
        Initialize();
    }
    init(buffer: NSMutableData)
    {
        m_data = buffer;
        hdr = NetworkHeader(buffer: buffer);
    }
    func Initialize()
    {
        <INIT_CODE>
    }

    <ENUMERATIONS>
    <ACCESSORS>

    // Reflection information
    /*
    public static let fields = [
        <REFLECTION>
    ]*/
};

// add our class to the dictionary
//MessageDictionary[<MSGID>] = <MSGNAME>;
